#include "ast3d.h"

/*
** static AstFixedpoint_t ComputeDistanceFactor( fov, width )
**
** This is a helper function used by the camera class to
** determine the horizontal and vertical distance factors used
** when doing perspective mappings.  Recall that the perspective
** transformation distance factor determines the field of view
** as follows:
**
**   tan( fov/2 ) = screen_width / ( 2 * df
**
**   thus solving for df gives us:
**
**   df = screen_width / ( 2 * tan( fov / 2 ) )
*/
static AstFixedpoint_t
ComputeDistanceFactor( unsigned fov, unsigned width )
{
AstFixedpoint_t half_width = AstConvertToFix( ( short ) width / 2 );
AstFixedpoint_t tangent = AstConvertToFix(
                            tan((fov/10.0)*M_PI/180.0));

   return ( AstFixDivFix( half_width, tangent ) );
}

/*
** inline int AstCamera::HProject( value, z )
**
** Projects a horizontal (x) value using the precomputed
** 'hdf' (horizontal distance factor) variable.  I've inlined
** this function for performance reasons, and since it doesn't
** hurt.
*/
inline int
AstCamera::HProject( AstFixedpoint_t v,AstFixedpoint_t z ) const
{
AstFixedpoint_t result = AstFixDivFix(AstFixMulFix(v,hdf),z);

   return AstConvertToInt( result ) + half_screen_width;
}

/*
** inline int AstCamera::VProject( value, z )
**
** Projects a vertical (y) value.  Note that since the VGA's
** coordinate system is the inverse of our camera system we must
** use a -z value for a projection (i.e. VGA +Y is down whereas
** our CCS +Y is up) to "flip" the Y value correctly.
*/
inline int
AstCamera::VProject(AstFixedpoint_t v,AstFixedpoint_t z ) const
{
AstFixedpoint_t result = AstFixDivFix(AstFixMulFix(v,vdf),-z);

   return AstConvertToInt( result ) + half_screen_height;
}

/*
** AstCamera::AstCamera( unsigned fov )
**
** This is the AstCamera class' constructor and is responsible
** for setting its various variables.  The only parameter to
** the constructor is the camera's desired field of view in
** tenths of degrees.  Typically a good value is 60 degrees.
** Larger values give a more "fish-eye" appearance.  The usable
** range is between 30 and 160 or so degrees.  You will
** probably cause some bad things to happen if you choose a
** value greater than about 175 degrees.
*/
AstCamera::AstCamera( unsigned fov_tenths_degrees )
{
   half_screen_width  = SCREEN_WIDTH / 2;
   half_screen_height = SCREEN_HEIGHT / 2;

   SetHFOV( fov_tenths_degrees, SCREEN_WIDTH );
   SetVFOV( fov_tenths_degrees, SCREEN_HEIGHT );
}

/*
** void AstCamera::ProjectPoints( s, d, n )
**
** This takes an array of "num_verts" vertices, projects them
** using the camera's projection values, and stores them in the
** "source_array" of screen vertices.
*/
void AstCamera::ProjectPoints( const AstVector *source_array,
                               AstScreenVertex *dest_array,
                               int num_vertices ) const
{
   for ( int i = 0; i < num_vertices; i++ ) {
      dest_array[i].X = HProject( source_array[i].X,
                                  source_array[i].Z );
      dest_array[i].Y = VProject( source_array[i].Y,
                                  source_array[i].Z );
   }
}

/*
** void AstCamera::SetHFOV( tenth_degrees, vport_width )
**
** Sets our horizontal field of view data based on our desired
** field of view and our view port width.
*/
void AstCamera::SetHFOV( unsigned tenth_degrees, int vwidth )
{
   hfov = tenth_degrees;
   hdf  = ComputeDistanceFactor( tenth_degrees / 2, vwidth );
}

/*
** void AstCamera::SetVFOV( tenth_degrees, vport_height )
**
** Sets our vertical field of view data based on our desired
** field of view and our view port height.
*/
void AstCamera::SetVFOV( unsigned tenth_degrees, int vheight )
{
   vfov = tenth_degrees;

   /*
   ** Vertical FOV automagically compensates for aspect ratio
   ** by multiplying against our desired aspect ratio 4:3.
   */
AstFixedpoint_t pre_aspect_ratio_vdf =
   ComputeDistanceFactor( tenth_degrees / 2, vheight );

   vdf = AstFixMulFix( AstConvertToFix( 4.0/3.0 ),
                       pre_aspect_ratio_vdf );
}

