#include "ast3d.h"

int NUM_ASTEROIDS = 12;

/*
** Private function declarations
*/
static void AstCreateAsteroids( void );
static void AstCreatePlayer( void );

/*
** Private variable declarations
*/
static AstTextureMap        asteroid_texmap;
static AstAsteroid         *asteroids;
static AstObjectDefinition *nasteroid_def;
static AstObjectDefinition *fasteroid_def;

/*
** Public variable declarations
*/
AstObjectDefinition *missile_def;

/*
** static void AstCreateAsteroids( void )
**
** This is a helper function that creates all of the variables
** associated with the creation of asteroids.
*/
static void AstCreateAsteroids( void )
{
   /*
   ** Load object definitions from disk
   */
   if ( ( nasteroid_def = new AstObjectDefinition( "OCTBALL.DEF", AstConvertToFix( 400 ) ) ) != 0 ) {
      if ( !nasteroid_def->IsValid() )
         AstFatal( "Error creating nasteroid_def!" );
   }
   if ( ( fasteroid_def = new AstObjectDefinition( "FAR_AST.DEF", AstConvertToFix( 1 ) ) ) != 0 ) {
      if ( !fasteroid_def->IsValid() )
         AstFatal( "Error creating fasteroid_def!" );
   }
   if ( ( missile_def = new AstObjectDefinition( "MISSILE.DEF", AstConvertToFix( 1 ) ) ) != 0 ) {
      if ( !missile_def->IsValid() )
         AstFatal( "Error creating missile_def!" );
   }

   /*
   ** Create the asteroids
   */
   if ( ( asteroids = new AstAsteroid[NUM_ASTEROIDS] ) == 0 )
      AstFatal( "Out of memory allocating asteroids!" );
   for ( int i = 0; i < NUM_ASTEROIDS; i++ ) {

      asteroids[i].SetNearDefinition( nasteroid_def, AstConvertToFix( 20 ) );
      asteroids[i].SetFarDefinition( fasteroid_def, AstConvertToFix( 100 ) );

      asteroids[i].SetName( "ASTEROID" );
   }

   /*
   ** Load the texture map from disk -- this doesn't actually
   ** load a texture map, but creates one on the fly in
   ** memory.
   */
   if ( !asteroid_texmap.Load( "asteroid.tex" ) )
      AstFatal( "Failed to load asteroid.tex!" );

   /*
   ** Assign the texture map to the various object definitions
   */
   missile_def->AssignTextureMap(   &asteroid_texmap );
   nasteroid_def->AssignTextureMap( &asteroid_texmap );
   fasteroid_def->AssignTextureMap( &asteroid_texmap );
}

/*
** static void AstCreatePlayer( void )
**
** This is a helper function that allocates memory for and creates the
** player.
*/
static void AstCreatePlayer( void )
{
  if ( ( player = new AstPlayer ) == 0 ) {
     AstFatal( "Out of memory allocating player!" );
  }
}

/*
** void AstEnd( void )
**
** Shuts down and frees up asteroids related stuff.
*/
void AstEnd( void )
{
   if ( nasteroid_def ) {
      delete nasteroid_def;
      nasteroid_def = 0;
   }
   if ( fasteroid_def ) {
      delete fasteroid_def;
      fasteroid_def = 0;
   }

   if ( asteroids ) {
      delete []asteroids;
      asteroids = 0;
   }
   if ( inputdevice2d ) {
      delete inputdevice2d;
      inputdevice2d = 0;
   }
   if ( gameport ) {
      delete gameport;
      gameport = 0;
   }
   if ( player ) {
      delete player;
      player = 0;
   }
   AstLibEnd();
}

/*
** void AstGetOpts( int argc, const char *argv[] )
**
** This routine is responsible for parsing the command line
** arguments passed to AST3D and setting various configuration
** variables accordingly.
*/
void AstGetOpts( int argc, const char *argv[] )
{

   /*
   ** Iterate through the command line arguments
   */
   for ( int i = 1; i < argc; i++ ) {

      /*
      ** Selecting the mouse disables the joystick, and vice
      ** versa
      */
      if ( strcmp( argv[i], "-mouse" ) == 0 ) {
         config.mouse    = true;
         config.joystick = false;
      }
      else if ( strcmp( argv[i], "-joystick" ) == 0 ) {
         config.joystick = true;
         config.mouse    = false;
      }

      /*
      ** RCS and WCS can be enabled/disabled independently of
      ** the mouse and joystick.
      */
      else if ( strcmp( argv[i], "-wcs" ) == 0 ) {
         config.wcs = true;
      }
      else if ( strcmp( argv[i], "-rcs" ) == 0 ) {
         config.rcs = true;
      }

      /*
      ** Check for texture mapping
      */
      else if ( strcmp( argv[i], "-texture" ) == 0 ) {
         config.texturing = true;
      }

      /*
      ** Calculate the number of asteroids we want
      */
      else if ( strncmp( argv[i], "-n", 2 ) == 0 ) {
         NUM_ASTEROIDS = atoi( argv[i] + 2 );
      }
   }
}

/*
** void AstInitInputDevices( void )
**
** This sets up the various input devices as specified by the
** "config" structure.
*/
void AstInitInputDevices( void )
{
   /*
   ** Set up the game port if options dictate it
   */
   int gpmask = 0;

   if ( config.joystick ) gpmask |= gpStick1;
   if ( config.rcs      ) gpmask |= gpRCS;
   if ( config.wcs      ) gpmask |= gpWCS;
   if ( gpmask ) {
     if ( ( gameport = new AstGameport( gpmask ) ) == 0 )
        AstFatal( "Could not allocate gameport!" );
     gameport->Calibrate();
     if ( config.rcs )
        if ( ( rcs = new AstRCS( gameport ) ) == 0 )
           AstFatal( "Failed to allocate RCS!" );
     if ( config.wcs )
        if ( ( wcs = new AstWCS( gameport ) ) == 0 )
           AstFatal( "Failed to allocate WCS!" );
   }

   /*
   ** Initialize the mouse
   */
   if ( config.mouse ) {
      if ( ( ( inputdevice2d = new AstMouse ) == 0 ) ||
           ( !inputdevice2d->IsAvailable() ) )
      {
         AstFatal( "Failed to create AstJoystic!" );
      }
   }
   /*
   ** Initialize the joystick
   */
   else if ( config.joystick ) {
      if ( ( inputdevice2d = new AstJoystick( gameport ) ) == 0 )
         AstFatal( "Failed to create AstJoystick!" );
   }
}

/*
** void AstInit( int argc, const char *argv[] )
**
** This is the primary initialization function for the entire
** AST3D game.  Initialization involves procedures that must
** be executed only at program startup, including:
**
**   input device initialization
**   graphics mode setting
**   creating objects (including the player)
**   allocating memory
**   inserting objects into the world
*/
void AstInit( int argc, const char *argv[] )
{
   /*
   ** Configure AST3D based on our command line parameters
   */
   AstGetOpts( argc, argv );

   /*
   ** Initialize the core AST3D routines
   */
   AstLibInit();

   /*
   ** Create and allocate our object definitions and objects
   */
   AstCreateAsteroids();
   AstCreatePlayer();

   /*
   ** Set up and create our input devices
   */
   AstInitInputDevices();

   /*
   ** Start up the graphics mode
   */
   vga->StartGraphics();

   /*
   ** Put the player into the world
   */
   world.Add( *player );

   /*
   ** Stick the asteroids into the world
   */
   for ( int i = 0; i < NUM_ASTEROIDS; i++ ) {
      world.Add( asteroids[i] );
   }
}

