#include "ast3d.h"

/*
** AstPlayer::AstPlayer( void )
**
** This is the AstPlayer class' constructor.  It sets various
** flags and computes a default bounding sphere because otherwise
** the bounding sphere would be 0 (since the player isn't visible,
** no definition is present).
*/
AstPlayer::AstPlayer( void ) : AstDrifter( "player" )
{
extern AstObjectDefinition *missile_def;

  bounding_sphere         = AstConvertToFix( 2 );
  bounding_sphere_squared = 4;

  missile.SetNearDefinition( missile_def, AstConvertToFix( 40 ) );
  missile.SetFarDefinition( missile_def, AstConvertToFix( 200 ) );
  world.Add( missile );

  SetFlag( ofSolid, true );
}

/*
** void AstPlayer::CheckInputDevices( void )
**
** Self-explanatory.
**
*/
void AstPlayer::CheckInputDevices( void )
{
   Check2DInputDevice();
   CheckThrustMaster();
   CheckKeyboard();
}

/*
** void AstPlayer::Check2DInputDevice( void )
**
** Checks the system default 2d input device (either a mouse
** or a joystick) for input.
*/
void AstPlayer::Check2DInputDevice( void )
{
   /*
   ** Make sure a 2D input device is being used
   */
   if ( inputdevice2d != 0 )
   {
      const float dead_zone = .15;

      /*
      ** Retrieve the values from the inputdevice
      */
      AstFixedpoint_t xinput = inputdevice2d->GetX();
      AstFixedpoint_t yinput = inputdevice2d->GetY();

      /*
      ** Check to make sure that X is outside the "dead zone"
      */
      if ( xinput > AstConvertToFix( dead_zone ) ||
           xinput < AstConvertToFix( -dead_zone ) )
      {
         /*
         ** If the keyboard shift key is pressed we rotate around the
         ** Z axis as opposed to the Y axis
         */
         if ( keyboard->Shift() )
         {
         AstFixedpoint_t amt = AstFixMulFix( inputdevice2d->GetX(), AstConvertToFix( 60 ) );

            RotateAroundZ( AstConvertToInt( -amt ) );
         }
         else
         {
            RotateAroundY( ( unsigned short ) AstConvertToInt(
                             AstFixMulFix( inputdevice2d->GetX(),
                                           AstConvertToFix( 40 ) ) ) );
         }
      }

      /*
      ** Make sure that the Y value is outside the "dead zone"
      */
      if ( yinput > AstConvertToFix( dead_zone ) ||
           yinput < AstConvertToFix( -dead_zone ) )
      {
        RotateAroundX( AstConvertToInt( -AstFixMulFix( inputdevice2d->GetY(),
                                                       AstConvertToFix( 40 ) ) ) );
      }

      /*
      ** Only use button 1 for thrust if the WCS is not being
      ** used.
      */
      if ( !config.wcs )
      {
         if ( inputdevice2d )
         {
            if ( inputdevice2d->GetButtons() & 1 )
               Thrust();
         }
      }

      /*
      ** Use button 2 to fire
      */
      if ( inputdevice2d )
      {
         if ( inputdevice2d->GetButtons() & 2 )
         {
            Fire();
         }
      }
   }
}

/*
** void AstPlayer::CheckKeyboard( void )
**
** Determines if the keyboard has a key waiting and acts upon
** any input if present.  The escape key will stop the program
** and the space key will halt the ship from drifting.
*/
void AstPlayer::CheckKeyboard( void )
{
int key;

   if ( ( key = keyboard->GetKey() ) != KbdNoKey )
   {
      switch ( key )
      {
         case KbdRightArrow:
            RotateAroundZ( -100 );
            break;
         case KbdUpArrow:
            RotateAroundX( 30 );
            break;
         case KbdDownArrow:
            RotateAroundX( -30 );
            break;
         case KbdLeftArrow:
            RotateAroundZ( 100 );
            break;
         case KbdEsc:
            world.Stop();
            break;
         case KbdSpace:
            Stop();
            break;
      }
   }
}

/*
** void AstPlayer::CheckThrustMaster( void )
**
** Checks any ThrustMaster Cockpit devices, such as the rudder
** control system or weapons control system, for input, assuming
** they are attached.
*/
void AstPlayer::CheckThrustMaster( void )
{
   if ( config.rcs )
   {
      RotateAroundY( AstConvertToInt( AstFixMulFix( rcs->GetValue(), AstConvertToFix( 10 ) ) ) );
   }
   if ( config.wcs )
   {
      AstFixedpoint_t adjusted_wcs = wcs->GetValue();

      adjusted_wcs -= AstConvertToFix( .5 );

      MoveAlongZ( AstFixMulFix( adjusted_wcs, AstConvertToFix( .3 ) ) );
   }
}

/*
** void AstPlayer::DoSomething( void )
**
** This is the AstPlayer class' DoSomething() function -- the
** player is responsible for handling input and drifting,
** if the weapon control system is not being used.  This also
** limits the player's speed by initiating a friction based
** braking action.
*/
void AstPlayer::DoSomething( void )
{
   CheckInputDevices();

   if ( !config.wcs )
      Drift();
   velocity *= AstConvertToFix( .90 );
}

/*
** void AstPlayer::Fire( void )
**
** Fires a missile from the player's ship.
*/
void AstPlayer::Fire( void )
{
   /*
   ** Make sure that no missiles are already in flight
   */
   if ( !missile.InFlight() )
   {
      /*
      ** Set the missile's initial velocity and position to correspond
      ** with those of the player
      */
      AstVector mvel;
      AstVector start( 0, 0, AstConvertToFix( 4 ) );
      AstVector vf( 0, 0, AstConvertToFix( 1 ) );
      AstMatrix3x3 rm;

      local_matrix.GetRotationMatrix( rm );

      vf    *= rm;
      start *= rm;

      mvel = velocity + vf;

      missile.MoveTo( GetPositionWCS() + start );
      missile.SetVelocity( mvel );
      missile.Fly();
   }
}

/*
** void AstPlayer::Thrust( void )
**
** Pressing button 1 of the default 2d input device causes
** the player's ship to increase velocity in its current
** facing.  This value is added to the ship's velocity
** vector.
*/
void AstPlayer::Thrust( void )
{
const AstFixedpoint_t speed_factor = AstConvertToFix( .2 );
AstVector v( 0, 0, AstConvertToFix( 1 ) );
AstMatrix3x3 orientation;

   local_matrix.GetRotationMatrix( orientation );
   v *= orientation;
   v *= speed_factor;
   velocity += v;
}

/*
** void AstPlayer::YouHit( target )
**
** This is the AstPlayer's class virtual YouHit() function,
** which is called whenever the player hits something such
** as an asteroid.  Right now when the player hits anything
** it stops moving.
*/
void AstPlayer::YouHit( AstObject * )
{
   Stop();
}
