#include "ast3d.h"

AstVGA   *vga;

/*
** Constants used by the VGA class
*/
#ifdef __BORLANDC__
static char * screen = ( char * ) MK_FP( 0xa000, 0 );
#else
static char * screen = ( char * ) 0xA0000;
#endif

const int    VIDEO_INTERRUPT = 0x10;
const size_t SCREEN_SIZE = SCREEN_WIDTH * SCREEN_HEIGHT;

/*
** AstVGA::AstVGA( void )
**
** This is the constructor for the AstVGA class.
*/
AstVGA::AstVGA( void )
{
   frames = 0;

   /*
   ** Allocate memory for the offscreen drawing buffer
   */
   if ( ( drawbuffer = new char[SCREEN_SIZE] ) == 0 )
      AstFatal( "Could not allocate enough memory for draw buffer!" );

   /*
   ** Load the palette from disk
   */
   if ( ( palette = new AstPalette( "VGA.PAL" ) ) != 0 ) {
      if ( !palette->IsValid() ) {
         AstFatal( "Invalid palette VGA.PAL" );
      }
   }
   else {
      AstFatal( "Could not allocate palette" );
   }
}

/*
** AstVGA::~AstVGA()
**
** This is the destructor for the AstVGA class, responsible for freeing
** memory associated with the drawbuffer and palette.
*/
AstVGA::~AstVGA()
{
   delete []drawbuffer;
   drawbuffer = 0;
   delete palette;
   palette = 0;
}

/*
** void AstVGA::BlitPage( void )
**
** This routine is responsible for copying the memory in the offscreen
** drawing buffer onto the screen.  It has been implemented using
** the slower memcpy() C-library function call for the same reasons
** that AstVGA::HLine() was implemented using memset().  The same
** caveats for AstVGA::HLine() in regards to performance also apply
** to AstVGA::BlitPage().
*/
void AstVGA::BlitPage( void )
{
   memcpy( screen, drawbuffer, SCREEN_SIZE );
   frames++;
}

/*
** void AstVGA::ClearScreen( unsigned char color )
**
** This routine resets the off-VGA display memory to the given color.
*/
void AstVGA::ClearScreen( unsigned char color )
{
   memset( drawbuffer, color, SCREEN_SIZE );
}

/*
** void AstVGA::Hline( x1, x2, y, color )
**
** This routine draws a horizontal line from x1 to x2 on scan line
** y in the given color.  It has been implemented using the relatively
** slow memset() C-library function call because the only alternative
** is either an inline assembly routine (which is Borland specific) or
** a pure assembly routine, requiring special code for Watcom and
** Borland.  This method is both more portable and easier to
** understand.  For a commercial application a specific assembly
** routine coded should probably be written.
*/
void AstVGA::HLine( int x1, int x2, int y, unsigned char color )
{
   /*
   ** Make sure Y value is legit
   */
   if ( y >= 0 && y < SCREEN_HEIGHT )
   {
      /*
      ** Make sure x1 < x2
      */
      if ( x1 > x2 )
      {
         int tmp = x1;
         x1 = x2;
         x2 = tmp;
      }

      /*
      ** Make sure at least part of the line is on the screen
      */
      if ( !( x1 < 0 && x2 < 0 ) &&
           !( x1 > SCREEN_WIDTH - 1 && x2 > SCREEN_WIDTH - 1 ) )
      {
         /*
         ** Clamp right and left edge values
         */
         if ( x1 < 0 )
            x1 = 0;
         if ( x2 > SCREEN_WIDTH - 1 )
            x2 = SCREEN_WIDTH - 1;

         /*
         ** Render the line
         */
         memset( drawbuffer + y * SCREEN_WIDTH + x1, color, x2 - x1 + 1 );
      }
   }
}

/*
** void AstVGA::TMHLine( x1, x2, u1, u2, v1, v2, y )a
**
** This function renders a texture mapped scan line to the offscreen
** buffer using the given x, y, u, and v coordinates.
*/
void AstVGA::TMHLine( AstFixedpoint_t x1, AstFixedpoint_t x2,
                      AstFixedpoint_t u1, AstFixedpoint_t u2,
                      AstFixedpoint_t v1, AstFixedpoint_t v2,
                      int y )
{
   /*
   ** Make sure Y value is legit
   */
   if ( y >= 0 && y < SCREEN_HEIGHT )
   {
      /*
      ** Make sure x1 < x2
      */
      if ( x1 > x2 )
      {
         AstFixedpoint_t tmp;

         tmp = x1;
         x1 = x2;
         x2 = tmp;

         tmp = u1;
         u1 = u2;
         u2 = tmp;

         tmp = v1;
         v1 = v2;
         v2 = tmp;
      }

      /*
      ** Make sure at least part of the line is on the screen
      */
      if ( !( x1 < 0 && x2 < 0 ) &&
           !( x1 > AstConvertToFix( SCREEN_WIDTH - 1 ) && x2 > AstConvertToFix( SCREEN_WIDTH - 1 ) ) )
      {

         AstFixedpoint_t cur_u, cur_v;
         AstFixedpoint_t du, dv;
         AstFixedpoint_t linewidth = x2 - x1;
         AstFixedpoint_t uwidth    = u2 - u1;
         AstFixedpoint_t vwidth    = v2 - v1;

         /*
         ** Clip against the left edge of the screen
         */
         if ( x1 < 0 )
         {
            AstFixedpoint_t ratio_clipped = AstFixDivFix( -x1, linewidth );

            /*
            ** Adjust u and v accordingly
            */
            u1 = u1 + AstFixMulFix( uwidth, ratio_clipped );
            v1 = v1 + AstFixMulFix( vwidth, ratio_clipped );

            x1 = 0;
         }

         /*
         ** Clip against the right edge of the screen
         */
         if ( x2 > AstConvertToFix( SCREEN_WIDTH - 1 ) )
         {
            AstFixedpoint_t ratio_clipped = AstFixDivFix( x2 - AstConvertToFix( SCREEN_WIDTH ), linewidth );

            /*
            ** Adjust u and v accordingly
            */
            u2 = u2 - AstFixMulFix( uwidth, ratio_clipped );
            v2 = v2 - AstFixMulFix( vwidth, ratio_clipped );

            x2 = AstConvertToFix( SCREEN_WIDTH - 1 );
         }

         /*
         ** Calculate width values
         */
         linewidth = x2 - x1;
         uwidth    = u2 - u1;
         vwidth    = v2 - v1;

         /*
         ** Make sure there is something to render
         */
         if ( AstConvertToInt( linewidth ) == 0 )
            return;

         /*
         ** Determine du and dv as ratios of texture width to linewidth
         */
         du = AstFixDivFix( uwidth, linewidth );
         dv = AstFixDivFix( vwidth, linewidth );

         /*
         ** Start off variables
         */
         cur_u = u1;
         cur_v = v1;

         /*
         ** Get a pointer to the first pixel we're rendering too
         */
         char *drawbuf = drawbuffer + y * SCREEN_WIDTH + AstConvertToInt( x1 );

         /*
         ** Iterate through the list
         */
         for ( AstFixedpoint_t x = x1; x < x2; x += AstConvertToFix( 1 ) )
         {
            *drawbuf = (*current_map)[AstConvertToInt( cur_u )][AstConvertToInt( cur_v )];
            cur_u += du;
            cur_v += dv;
            drawbuf++;
         }
      }
   }
}

/*
** void AstVGA::SetPalette( unsigned char palette[256][3] )
**
** Sets the VGA default color palette.  The values passed are assumed
** to be 18-bit R,G,B triplets.  The most significant 2-bits of each
** color value are ignored, i.e. the highest value per triplet should
** be (63,63,63).
** It is important that when the palette is created that word alignment
** is disabled.  The data MUST be byte aligned otherwise the palette
** setting routine called by int 0x10 will read incorrect values.
*/
void AstVGA::SetPalette( AstPalette &palette )
{
REGS  regs;
SREGS sregs;

   /*
   ** Set the "regs" registers to the appropriate values
   */
   REGISTERWORD( regs, ax ) = 0x1012;
   REGISTERWORD( regs, bx ) = 0;
   REGISTERWORD( regs, cx ) = 256;

   /*
   ** Borland 16-bit real-mode requires only setting DX to the
   ** offset of palette.data.  However, Watcom 32-bit requires
   ** setting EDX to the offset (since the offset may be larger
   ** than 16-bits).  Also, the example for FP_SEG() under
   ** Watcom sets DS = FS = GS = FP_SEG( &regs ) so I'm doing
   ** it here to maintain consistency.
   */
#ifdef __BORLANDC__
   REGISTERWORD( regs, dx ) = FP_OFF( palette.data );
#else
   regs.x.edx = FP_OFF( palette.data );
   sregs.ds = sregs.fs = sregs.gs = FP_SEG( &regs );
#endif

   /*
   ** ES is set to the segment of palette.data
   */
   sregs.es = FP_SEG( palette.data );

   INT86X( VIDEO_INTERRUPT, &regs, &regs, &sregs );
}

/*
** void AstVGA::SetupDefaultPalette( void )
**
** This routine gives us a nice palette of 16 shades of 16 colors.
*/
void AstVGA::SetupDefaultPalette( void )
{
   SetPalette( *palette );
}

/*
** void AstVGA::StartGraphics
**
** This function puts the VGA card into graphics mode 0x13, 320x200x256
** with one page.  It does this by issuing an interrupt 0x10 with the
** AX register set to 0x13.
**   After doing this it sets up the default palette, clears the screen
** to black, and then blits out the (empty) page to clear the framebuffer
** on the card.
*/
void AstVGA::StartGraphics( void )
{
REGS regs;

   REGISTERWORD( regs, ax ) = 0x13;
   INT86( VIDEO_INTERRUPT, &regs, &regs );
   SetupDefaultPalette();
   ClearScreen( 0 );
   BlitPage();
}

/*
** void AstVGA::StopGraphics( void )
**
** This function resets the VGA card to text mode by setting it to
** video mode 0x03.
*/
void AstVGA::StopGraphics( void )
{
REGS regs;

   REGISTERWORD( regs, ax ) = 0x03;
   INT86( VIDEO_INTERRUPT, &regs, &regs );
}

/*
** void AstVGA::TextureMappedTriangle( ... )
**
** This routine renders a texture mapped triangle out to the screen,
** using basically the same algorithm as the AstVGA::Triangle(...)
** routine but adjusting/calculating U, V coordinates also.
*/
void AstVGA::TextureMappedTriangle( const AstScreenVertex &a, const AstScreenVertex &b, const AstScreenVertex &c,
                                    const AstScreenVertex &ta, const AstScreenVertex &tb, const AstScreenVertex &tc,
                                    const AstTextureMap *map )
{
/*
** Cache variables locally
*/
int x1 = a.X;
int x2 = b.X;
int x3 = c.X;
int y1 = a.Y;
int y2 = b.Y;
int y3 = c.Y;
int u1 = ta.X;
int u2 = tb.X;
int u3 = tc.X;
int v1 = ta.Y;
int v2 = tb.Y;
int v3 = tc.Y;

int tmp;

   /*
   ** Sort our vertices from top to bottom
   */
   if ( y3 < y1 ) {
      tmp = y3;
      y3 = y1;
      y1 = tmp;
      tmp = x3;
      x3 = x1;
      x1 = tmp;

      tmp = v3;
      v3 = v1;
      v1 = tmp;
      tmp = u3;
      u3 = u1;
      u1 = tmp;
   }

   if ( y2 < y1 ) {
      tmp = y2;
      y2 = y1;
      y1 = tmp;
      tmp = x2;
      x2 = x1;
      x1 = tmp;

      tmp = v2;
      v2 = v1;
      v1 = tmp;
      tmp = u2;
      u2 = u1;
      u1 = tmp;
   }

   if ( y3 < y2 ) {
      tmp = y3;
      y3 = y2;
      y2 = tmp;
      tmp = x3;
      x3 = x2;
      x2 = tmp;

      tmp = v3;
      v3 = v2;
      v2 = tmp;
      tmp = u3;
      u3 = u2;
      u2 = tmp;
   }

   /*
   ** If the triangle has no horizontal edges then split it into
   ** two triangles that do.
   */
   if ( y2 != y1 && y3 != y1 && y3 != y2 ) {
      int             xn, un, vn;
      AstFixedpoint_t ratio;

      /*
      ** Determine new x, u, and n values
      **   xn = ( x3 - x1 ) * ( ( y2 - y1 ) / ( y3 - y1 ) ) + x1
      **   un = ( u3 - u1 ) * ( ( y2 - y1 ) / ( y3 - y1 ) ) + u1
      **   vn = ( v3 - v1 ) * ( ( y2 - y1 ) / ( y3 - y1 ) ) + u1
      */

      /*
      ** First precompute the ratio of ( y2 - y1 ) / ( y3 - y1 )
      */
      ratio = AstFixDivFix( AstConvertToFix( y2 - y1 ),
                            AstConvertToFix( y3 - y1 ) );

      xn =   AstConvertToInt( AstFixMulFix( AstConvertToFix( x3 - x1 ), ratio ) )
           + x1;
      un =   AstConvertToInt( AstFixMulFix( AstConvertToFix( u3 - u1 ), ratio ) )
           + u1;
      vn =   AstConvertToInt( AstFixMulFix( AstConvertToFix( v3 - v1 ), ratio ) )
           + v1;

      /*
      ** Draw the two subtriangles
      */
      if ( xn < x2 ) {
         FB_TMTriangle( x1, y1, u1, v1,
                        xn, y2, un, vn,
                        x2, y2, u2, v2, *map );
         FB_TMTriangle( x3, y3, u3, v3,
                        x2, y2, u2, v2,
                        xn, y2, un, vn, *map );
      }
      else {
         FB_TMTriangle( x1, y1, u1, v1,
                        x2, y2, u2, v2,
                        xn, y2, un, vn, *map );
         FB_TMTriangle( x3, y3, u3, v3,
                        x2, y2, u2, v2,
                        xn, y2, un, vn, *map );
      }
   }
   /*
   ** Top is flat
   */
   else if ( y1 == y2 ) {
      if ( x1 < x2 ) {
         FB_TMTriangle( x3, y3, u3, v3,
                        x2, y2, u2, v2,
                        x1, y2, u1, v1, *map );
      }
      else {
         FB_TMTriangle( x3, y3, u3, v3,
                        x1, y2, u1, v1,
                        x2, y2, u2, v2, *map );
      }
   }
   /*
   ** Bottom is flat
   */
   else if ( y2 == y3 ) {
      if ( x2 < x3 ) {
         FB_TMTriangle( x1, y1, u1, v1,
                        x2, y2, u2, v2,
                        x3, y3, u3, v3, *map );
      }
      else {
         FB_TMTriangle( x1, y1, u1, v1,
                        x3, y3, u3, v3,
                        x2, y2, u2, v2, *map );
      }
   }
}

/*
** void AstVGA::Triangle( a, b, c, color );
**
** This routine is the basic triangle routine core to the AstVGA
** class.  It works by splitting a triangle horizontally in half
** so that you have two flat sided triangles.  These two subtriangles
** are then drawn independently.
*/
void AstVGA::Triangle( const AstScreenVertex &a,
                       const AstScreenVertex &b,
                       const AstScreenVertex &c,
                       unsigned char color )
{
/*
** Cache variables locally
*/
int x1 = a.X;
int x2 = b.X;
int x3 = c.X;
int y1 = a.Y;
int y2 = b.Y;
int y3 = c.Y;
int tmp;

   /*
   ** Sort our vertices from top to bottom
   */
   if ( y3 < y1 ) {
      tmp = y3;
      y3 = y1;
      y1 = tmp;
      tmp = x3;
      x3 = x1;
      x1 = tmp;
   }

   if ( y2 < y1 ) {
      tmp = y2;
      y2 = y1;
      y1 = tmp;
      tmp = x2;
      x2 = x1;
      x1 = tmp;
   }

   if ( y3 < y2 ) {
      tmp = y3;
      y3 = y2;
      y2 = tmp;
      tmp = x3;
      x3 = x2;
      x2 = tmp;
   }

   /*
   ** If the triangle has no horizontal edges then split it into
   ** two triangles that do.
   */
   if ( y2 != y1 && y3 != y1 && y3 != y2 ) {
      int             xn;
      AstFixedpoint_t ratio;

      /*
      ** Determine new x value:
      **   xn = ( x3 - x1 ) * ( ( y2 - y1 ) / ( y3 - y1 ) ) + x1
      */

      /*
      ** Precompute the ratio of ( y2 - y1 ) / ( y3 - y1 )
      */
      ratio = AstFixDivFix( AstConvertToFix( y2 - y1 ),
                            AstConvertToFix( y3 - y1 ) );

      /*
      ** xn = ( ( x3 - x1 ) * ratio ) + x1
      */
      xn =   AstConvertToInt( AstFixMulFix( AstConvertToFix( x3 - x1 ), ratio ) )
           + x1;

      /*
      ** Draw the two subtriangles
      */
      if ( xn < x2 ) {
         FB_Triangle( x1, y1, xn, y2, x2, y2, color );
         FB_Triangle( x3, y3, x2, y2, xn, y2, color );
      }
      else {
         FB_Triangle( x1, y1, x2, y2, xn, y2, color );
         FB_Triangle( x3, y3, x2, y2, xn, y2, color );
      }
   }
   /*
   ** Top is flat
   */
   else if ( y1 == y2 ) {
      if ( x1 < x2 ) {
         FB_Triangle( x3, y3, x2, y2, x1, y2, color );
      }
      else {
         FB_Triangle( x3, y3, x1, y2, x2, y2, color );
      }
   }
   /*
   ** Bottom is flat
   */
   else if ( y2 == y3 ) {
      if ( x2 < x3 ) {
         FB_Triangle( x1, y1, x2, y2, x3, y2, color );
      }
      else {
         FB_Triangle( x1, y1, x3, y2, x2, y2, color );
      }
   }
}

/*
** void AstVGA::FB_Triangle( x1, y1, x2, y23, x3, , color )
**
** This routine renders a flat-bottomed triangle to the screen.  The
** reason the triangle must have a flat bottom is so that the delta
** X values for the left and right sides can then be guaranteed to
** never change over the entire span of the triangle, making the code
** simpler and faster.
*/
void AstVGA::FB_Triangle( int x1, int y1,
                          int x2, int y23,
                          int x3, int    ,
                          unsigned char color )
{
AstFixedpoint_t dx_left, dx_right;
AstFixedpoint_t curx_left, curx_right;
AstFixedpoint_t height = y23 - y1;

   /*
   ** Make sure that y23 - y1 != 0
   */
   if ( height == 0 )
      return;

   /*
   ** Determine the change in X (i.e. slope) for left and right sides
   */
   dx_left  = AstConvertToFix( x2 - x1 ) / height;
   dx_right = AstConvertToFix( x3 - x1 ) / height;

   curx_left = curx_right = AstConvertToFix( x1 );

   /*
   ** Draw each scan line from top to bottom
   */
   if ( y1 < y23 )  {
      for ( int y = y1; y <= y23; y++ ) {

        HLine( AstConvertToInt( curx_left ), AstConvertToInt( curx_right ), y, color );

        /*
        ** Adjust X coordinates
        */
        curx_left  += dx_left;
        curx_right += dx_right;
      }
   }
   else {
      for ( int y = y1; y >= y23; y-- ) {

        HLine( AstConvertToInt( curx_left ), AstConvertToInt( curx_right ), y, color );

        /*
        ** Adjust X coordinates
        */
        curx_left  -= dx_left;
        curx_right -= dx_right;
      }
   }
}

/*
** void AstVGA::FB_TMTriangle( x1, y1,  u1, v2,
**                             x2, y23, u2, v2,
**                             x3,    , u3, v3 )
**
** This is the texture mapped analog of AstVGA::FB_Triangle().
*/
void AstVGA::FB_TMTriangle( int x1, int y1,  int u1, int v1,
                            int x2, int y23, int u2, int v2,
                            int x3, int    , int u3, int v3,
                            const AstTextureMap &map )
{
AstFixedpoint_t dx_left, dx_right;
AstFixedpoint_t du_left, du_right;
AstFixedpoint_t dv_left, dv_right;
AstFixedpoint_t curx_left, curx_right;
AstFixedpoint_t curu_left, curu_right;
AstFixedpoint_t curv_left, curv_right;
AstFixedpoint_t height = y23 - y1;

   if ( height == 0 )
      return;

   SetCurrentMap( &map );

   /*
   ** Determine change in X, U, and V for the two sides
   */
   dx_left  = AstConvertToFix( x2 - x1 ) / height;
   dx_right = AstConvertToFix( x3 - x1 ) / height;
   du_left  = AstConvertToFix( u2 - u1 ) / height;
   du_right = AstConvertToFix( u3 - u1 ) / height;
   dv_left  = AstConvertToFix( v2 - v1 ) / height;
   dv_right = AstConvertToFix( v3 - v1 ) / height;

   curx_left = curx_right = AstConvertToFix( x1 );
   curu_left = curu_right = AstConvertToFix( u1 );
   curv_left = curv_right = AstConvertToFix( v1 );

   /*
   ** Draw each scan line from top to bottom
   */
   if ( y1 < y23 ) {
      for ( int y = y1; y <= y23; y++ ) {
         TMHLine( curx_left,
                  curx_right,
                  curu_left, curu_right,
                  curv_left, curv_right,
                  y );

        /*
        ** Adjust scan line values
        */
        curx_left  += dx_left;
        curx_right += dx_right;
        curu_left  += du_left;
        curu_right += du_right;
        curv_left  += dv_left;
        curv_right += dv_right;
      }
   }
   else {
      for ( int y = y1; y >= y23; y-- ) {
         TMHLine( curx_left,
                  curx_right,
                  curu_left, curu_right,
                  curv_left, curv_right,
                  y );

        /*
        ** Adjust scan line values
        */
        curx_left  -= dx_left;
        curx_right -= dx_right;
        curu_left  -= du_left;
        curu_right -= du_right;
        curv_left  -= dv_left;
        curv_right -= dv_right;
      }
   }
}

/*
** void AstVGA::VLine( x1, y1, y2, color )
**
** This routine draws a vertical line from y1 to y2 at row x1 with
** the given color.  This routine is rarely called by AST3D, so
** performance is not an issue.  This routine is probably the slowest
** of the graphics output routines compared to a similar implementation
** in assembly or inline assembly.  However, this implementation
** method is highly portable.
*/
void AstVGA::VLine( int x1, int y1, int y2, unsigned char color )
{
   /*
   ** Make sure the line is legit
   */
   if ( x1 >= 0 && x1 < SCREEN_WIDTH )
   {
      /*
      ** Make sure y1 < y2
      */
      if ( y1 > y2 )
      {
         int tmp = y1;
         y1 = y2;
         y2 = tmp;
      }

      /*
      ** Make sure part of the line is on the screen
      */
      if ( !( y1 < 0 && y2 < 0 ) &&
           !( y1 > SCREEN_HEIGHT - 1 && y2 > SCREEN_HEIGHT - 1 ) )
      {
         /*
         ** Clamp values vertically
         */
         if ( y1 < 0 )
            y1 = 0;
         if ( y2 > SCREEN_HEIGHT - 1 )
            y2 = SCREEN_HEIGHT - 1;

         /*
         ** Get pointer to the first pixel
         */
         char *dptr   = drawbuffer + y1 * SCREEN_WIDTH + x1;
         int   height = y2 - y1 + 1;

         /*
         ** Render
         */
         for ( int i = 0; i < height; i++ )
         {
            *dptr = color;
            dptr += SCREEN_WIDTH;
         }
      }
   }
}

